<?PHP
/**
 * MulitGet
 * Version 1.2
 * Copyright (C) 2010 All Rights Reserved.
 *
 * @Application : MultiGet.class.php
 * @Encode : 65001 Unicode ( UTF-8 NO BOM )
 * @Description : 多线程下载核心类，断点续传，状态控制，缓存控制...
 *						  需要SQLite及php_curl扩展
 * @Author : HK <hkyyang@qq.com>
 * @Update : 2010/12/18
 *					更新不支持断点的情况，在任务开始后转为单线程
 */

/**
 * ------------------------------ 方法列表 ------------------------------
 *
 * --@ static
 *	task_status					任务状态更新
 *	task_header					获取header
 *	task_remove					删除任务
 *	task_standby					队列等待
 *	task_pause					任务暂停
 *	task_main_select			用url或id查询主任务表
 *	parseMac						获取Resource类型的字符串
 *	multi_thread_callback		线程数据包返回处理
 *
 *	--@ public
 *	task_create					新建任务
 *	task_continue				继续任务
 *
 *	--@ private
 *	task_start						任务开始
 *	multi_add_handle			添加线程
 *	multi_exec_handle			执行线程
 *	task_finish						网络流结束或异常中断
 *	multi_update					线程进度更新
 *	task_cache					缓存控制
 *	task_complete				任务完成
 *	task_failure					任务失败
 *
 */


class MultiGet {

	// @ 统一缓存根结点名称 MG

    public function __construct( $dbName = null, $autoCreate = true, $pconnect = false ){

		if ( intval( phpversion() ) < 5 ) {
           exit( 'Error: Need a higher version.' );
        } else if( ! extension_loaded( 'curl' ) ) {
			exit( "Error: Unable to load extension \"php_curl.". PHP_SHLIB_SUFFIX ."\"." );
		}

	}

	/**
	 * 新建一个任务
	 *
	 * @access public
	 * @param String $url
	 * @param Integer $setThread : Default 1 任务线程数
	 * @param Integer $status : Default TASK_INIT 状态: TASK_INIT | TASK_STANDBY | TASK_PAUSE
	 * @param Integer $level 优先级别
	 * @param String $rename 保存重命名
	 * @return Boolean
	 */
	public function task_create( $url, $setThread = 1, $status = TASK_INIT, $level = 0, $rename = null ) {

		// URL 解析
		$url = parseURL( $url );

		if( ! in_array( $status, array( TASK_INIT, TASK_STANDBY, TASK_PAUSE ) ) ) {
			log_write( '新建任务的状态不正确, ' . $status );
			exit;
		}

		$info = self :: task_header( $url );
		if( $info === false ) return false;

		// 强制单线程
		$thread = $info[ 'http' ] != 0 ? $setThread : 1;
		// 超过最大线程限制
		$thread = $thread > Cache :: get( 'CONFIG.TASK_THREAD_MAX' ) ? Cache :: get( 'CONFIG.TASK_THREAD_MAX' ) : $thread;
		// 未达到多线程最小大小, 强制单线程
		if( $info[ 'size' ] < Cache :: get( 'CONFIG.TASK_MULTI_SIZE' ) ) {
			$thread = 1;
		}
		// 计算子线程分段
		$range = ceil( $info[ 'size' ] / $thread );
		$threadArr = array();
		for( $i = 0; $i < $thread; $i ++ ) {
			if( ( $i + 1 ) == $thread && $thread > 1 ) {
				$threadArr[ $i ] = array(
					'start' => $threadArr[ $i - 1 ][ 'end' ] + 1,
					'end' => $info[ 'size' ] - 1
				);
			} else if( $i < 1 ) {
				$threadArr[ $i ] = array(
					'start' => 0,
					'end' => $range - 1
				);
			} else {
				$threadArr[ $i ] = array(
					'start' => $threadArr[ $i - 1 ][ 'end' ] + 1,
					'end' => ( $i + 1 ) * $range
				);
			}
		}

		$name = $rename ? $rename : basename( $url );
		$file_name = Cache :: get( 'CONFIG.TASK_SAVE_DIR' ) . $name;

		// 任务主信息
		Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( "url='{$url}'" ) -> delete();
		Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> insert( array(
			'url' => $url,
			'name' => $name,
			'file_name' => $file_name,
			'size' => $info[ 'size' ],
			'status' => $status,
			'thread' => $thread,
			'set_thread' => $setThread,
			'http' => $info [ 'http' ],
			'level' => $level,
			'create_time' => time()
		) );
		$main_id = Cache :: get( 'SQLITE' ) -> lastInsertID;

		// 任务子线程
		Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> where( "uid={$main_id}" ) -> delete();
		$thread_info = $_tmp = array();
		for( $i = 0; $i < $thread; $i ++ ) {
			$_tmp = array(
				'uid' => $main_id,
				'start' => $threadArr[ $i ][ 'start' ],
				'end' => $threadArr[ $i ][ 'end' ],
				'pointer' => $threadArr[ $i ][ 'start' ],
				'level' => ( $i + 1 ),
			);
			$thread_info[] = $_tmp;
		}
		Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> insert( $thread_info );

		return true;

	}

	/**
	 * 继续/开始一个已经存在的任务
	 *
	 * @access public
	 * @param String | Integer $urlORid
	 * @return Boolean
	 */
	public function task_continue( $urlORid ) {

		$task_num = Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( 'status=' . TASK_START ) -> count();
		// 非再次循环并且超过最大进行任务数，转入队列等待
		if( $task_num >= Cache :: get( 'CONFIG.TASK_START_MAX' ) && is_null( Cache :: get( 'STATIC.REPEAT_NUM' ) ) ) {
			return $this -> task_standby( $urlORid );
		}

		// 当前任务再循环次数
		Cache :: set( 'STATIC.REPEAT_NUM', 1, '+=' );
		// 达到最大重试限制，判定此任务下载失败
		if( Cache :: get( 'STATIC.REPEAT_NUM' ) >= Cache :: get( 'CONFIG.TASK_REPEAT' ) ) {
			$this -> task_failure( $urlORid );
			return false;
		// 第一次循环时，记录时间用于计算任务用时
		} else if( Cache :: get( 'STATIC.REPEAT_NUM' ) == 1 ) {
			Cache :: set( 'RUNTIME', microtime( true ) );
		}

		$task = self :: task_main_select( $urlORid );

		// 错误: 任务不存在
		if( count( $task ) < 1 ) {
			return false;
		}

		// 查询子线程信息
		$thread_list = Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> where( "uid={$task[ 0 ][ 'id' ]}" ) -> order( 'level ASC' ) -> select();
		// 任务已完成
		if( count( $thread_list ) == 0 && $task[ 0 ][ 'size' ] == $task[ 0 ][ 'data' ] ) {
			return $this -> task_complete( $urlORid );
		// 错误: 线程数量不一致
		} else if( count( $thread_list ) != $task[ 0 ][ 'thread' ] ) return false;

		//验证每个一子线程
		$is_init = true;
		foreach( $thread_list as $thread ) {
			if( $thread[ 'data' ] > 0 ) {
				$is_init = false;
				break;
			}
		}

		// 错误: 任务是继续状态，但临时文件已经不存在
		if( ! $is_init && ! file_exists( $task[ 0 ][ 'file_name' ] . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ) ) ) {
			log_write( '临时文件不存在' );
			exit;
		// 警告: 任务初始化但临时文件已经存在
		} else if( $task[ 0 ][ 'status' ] == TASK_INIT && file_exists( $task[ 0 ][ 'file_name' ] . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ) ) ) {

			// 错误: 任务未开始情况下，子线程已有活动记录
			if( $is_init ) return false;
			// 错误: 临时文件大小与任务大文件小不一致
			if( filesize( $task[ 0 ][ 'file_name' ] . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ) ) != $task[ 0 ][ 'size' ] ) {
				return false;
			}

		}

		// 缓存变量
		Cache :: set( 'MG', array(
			'URL' => $task[ 0 ][ 'url' ],
			'SIZE' => $task[ 0 ][ 'size' ],
			'MAIN_ID' => $task[ 0 ][ 'id' ],
			'FILE_NAME' => $task[ 0 ][ 'file_name' ],
			'TEMP_FILE' => $task[ 0 ][ 'file_name' ] . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ),
			'THREAD_NUM' => $task[ 0 ][ 'thread' ], // 当前任务实际线程数
			'THREAD_CHECK' => false, // 多线程验证
			'THREAD_ONE' => false, // 只使用单线程
			'THREAD_CACHE_SIZE_LIMIT' => floor( sizeFormat( Cache :: get( 'CONFIG.TASK_MEM_LIMIT' ), 'M', 'B' ) / $task[ 0 ][ 'thread' ] ) - 1024 * 1024 * 2
		) );

		// 创建临时文件分配任务大小
		$cr = createBytes( Cache :: get( 'MG.TEMP_FILE' ), Cache :: get( 'MG.SIZE' ), $is_init );
		if( false === $cr ) {
			$this -> task_failure( $task[ 0 ][ 'id' ] );
			return false;
		}

		// 标记任务状态为开始下载
		Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( "id=" . Cache :: get( 'MG.MAIN_ID' ) ) -> update( array( 'status' => TASK_START ) );
		Cache :: set( 'MG.STATUS.START', true );
		$this -> task_start();
		return false;
	}

	/**
	 * 开始任务，初始化子线程
	 *
	 * @access private
	 * @param $
	 * @return Boolean
	 */
	private function task_start() {

		// 任务子线程，有可能是继续任务，跳过已经完成的线程，以指针位置判断
		$thread_list = Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> where( 'uid=' . Cache :: get( 'MG.MAIN_ID' ) . 'AND thread_list.pointer < thread_list.end' ) -> select();
		// 任务不存在或已完成
		if( count( $thread_list ) < 1 ) return false;

		$MultiHandle = curl_multi_init();
		Cache :: set( 'MG.MULTI_HANDLE', $MultiHandle );

		// 循环添加子线程
		foreach ( $thread_list as $thread ) {
			$CurlHandle = $this -> multi_add_handle ( $MultiHandle, $thread );
			if( $CurlHandle === false ) continue;
			Cache :: set( 'MG.CURL_HANDLE.*', $CurlHandle );
		}
		// 执行任务
		$this -> multi_exec_handle( $MultiHandle );
		return false;
	}

	/**
	 * 开始任务，初始化子线程
	 *
	 * @access private
	 * @param Resource $MultiHandle
	 * @param Array $thread 子线程信息
	 * @return Resource
	 */
	private function multi_add_handle( $MultiHandle, $thread_info ) {

		// 如果指针位置大于结束点，线程已完成
		// 通常线程正常完成的情况下，指针位置为结束点+1
		if( $thread_info[ 'pointer' ] < $thread_info[ 'end' ] ) {
			$pointer = $thread_info[ 'pointer' ] >= $thread_info[ 'start' ] ? $thread_info[ 'pointer' ] : $thread_info[ 'start' ];
		} else {
			return false;
		}

		$CurlHandle = curl_init( Cache :: get( 'MG.URL' ) );
		$mac = self :: parseMac( $CurlHandle );

		// 打开线程文件指针，设置指针偏移
		$FileHandle = fopen( Cache :: get( 'MG.TEMP_FILE' ), 'r+b' );

		fseek( $FileHandle, $pointer );

		// 缓存文件指针、线程ID、指针位置、指针结束点
		Cache :: set( 'CACHE.THREAD.' . $mac, array(
			'FILE_HANDLE' => $FileHandle,
			'ID' => $thread_info[ 'id' ],
			'POINTER' => $pointer,
			'END' => $thread_info[ 'end' ]
		) );
		$rangeBytes = $pointer . '-' . $thread_info[ 'end' ];

		/* ------------------------------- 待完善 -------------------------------
		// 对每个线程使用不同的代理
		curl_setopt($ch, CURLOPT_PROXY, '10.0.0.1:8080');
		curl_setopt($ch, CURLOPT_PROXYUSERPWD,'user:pass');
		 * ------------------------------- 待完善 -------------------------------*/

		curl_setopt( $CurlHandle, CURLOPT_HTTPHEADER, array( 'Range: Bytes=' . $rangeBytes ) );
		curl_setopt( $CurlHandle, CURLOPT_FILE, $FileHandle );
		curl_setopt( $CurlHandle, CURLOPT_WRITEFUNCTION, array( 'MultiGet', 'multi_thread_callback' ) );

		/* ------------------------------------------------------------------------TIMEOUT----------------------------------------------------------------------
			如果指定了一个数值(秒)，绝对会在指定秒后断开，关闭可以提高性能。
			根据程序功能会进入再次循环，并被处理为断点续传，还要经过一些不必要的验证，
			并且STATIC.REPEAT_NUM值+1，当重复次数达到限制，此任务被判定为下载失败。
		   ------------------------------------------------------------------------------------------------------------------------------------------------------------*/
		if( Cache :: get( 'CONFIG.TASK_TIMEOUT' ) > 0 ) {
			curl_setopt( $CurlHandle, CURLOPT_TIMEOUT, Cache :: get( 'CONFIG.TASK_TIMEOUT' ) );
		}

		curl_multi_add_handle( $MultiHandle, $CurlHandle );

		return $CurlHandle;
	}

	/**
	 * 主要执行方法，循环线程
	 *
	 * @access private
	 * @param Resource $MultiHandle
	 * @return Void
	 */
	private function multi_exec_handle( $MultiHandle ) {
		$active = false;
		do {
			$ch = curl_multi_exec( $MultiHandle, $active );
		} while( $ch == CURLM_CALL_MULTI_PERFORM && Cache :: get( 'MG.STATUS.START' ) );

		while( $active && $ch == CURLM_OK && Cache :: get( 'MG.STATUS.START' ) && Cache :: get( 'CACHE.NULL_DATA_NUM' ) <= Cache :: get( 'CONFIG.TASK_NULL_DATA_LIMIT' ) ) {
			if( curl_multi_select( $MultiHandle ) != -1 && Cache :: get( 'MG.STATUS.START' ) ) {
				do {
					$ch = curl_multi_exec( $MultiHandle, $active );
					$this -> task_cache();
					self :: events( true );
				} while( $ch == CURLM_CALL_MULTI_PERFORM && Cache :: get( 'MG.STATUS.START' ) && Cache :: get( 'CACHE.NULL_DATA_NUM' ) <= Cache :: get( 'CONFIG.TASK_NULL_DATA_LIMIT' ) );
			}
		}

		// 到达这一步的可能情况
		// 1. 用户手动停止任务
		// 2. 连续CONFIG.TASK_NULL_DATA_LIMIT个空数据包
		// 3. 任务已删除
		// 4. 设置了一个TIMEOUT值
		// 5. 任务完成

		$this -> task_finish();
	}

	/**
	 * 处理事件列表
	 *
	 * @static
	 * @access public
	 * @return Void
	 */
	static public function events( $is_loop = false ) {
		$list = Cache :: get( 'SQLITE' ) -> table( 'events_list' ) -> order( 'create_time ASC' ) -> select();
		for( $i = 0; $i < count( $list ); $i ++ ) {
			if( $list[ $i ][ 'exec_time' ] > 0 && time() < $list[ $i ][ 'exec_time' ] ) continue;
			switch( $list[ $i ][ 'type' ] ) {
				case 'remove' :
					// 删除的任务既是当前正在进行的任务
					if( $list[ $i ][ 'id' ] == Cache :: get( 'MG.MAIN_ID' ) && $is_loop == true ) {
						// 先中断接收数据包，此时无法删除任务
						// 为减少出错，只在资源指针关闭状态下处理
						Cache :: set( 'MG.STATUS.START', false );
						// 直接跳出，等待下一次的事件处理
						return false;
					} else {
						self :: task_remove( $list[ $i ][ 'id' ], $list[ $i ][ 'event' ] == '1' ? true : false );
						self :: removeEvent( $list[ $i ] );
					}
					break;
				case 'truncate' :
					if( $is_loop == true ) {
						Cache :: set( 'MG.STATUS.START', false );
						return false;
					} else {
						Cache :: get( 'SQLITE' ) -> truncate( 'task_list' );
						Cache :: get( 'SQLITE' ) -> truncate( 'thread_list' );
						$file = glob( Cache :: get( 'CONFIG.TASK_SAVE_DIR' ) . '*' . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ) );
						if( count( $file ) ) {
							foreach( $file as $f ) {
								@unlink( $f );
							}
						}
						self :: removeEvent( $list[ $i ] );
					}
					break;
			}
		}
		return false;
	}

	static public function removeEvent( $event ) {
		if( count( $event ) < 1 ) return false;
		return Cache :: get( 'SQLITE' ) -> table( 'events_list' ) -> where( "type='{$event[ 'type' ]}' AND create_time={$event[ 'create_time' ]}" ) -> delete();
	}

	/**
	 * 异常或正常网络中断，判断完成状态，再次循环
	 *
	 * @access private
	 * @return Boolean
	 */
	private function task_finish() {

		// 为了指针和实际完成状态同步，强制写缓存
		if( true != Cache :: get( 'MG.THREAD_ONE' ) ) $this -> task_cache( true );

		// 关闭所有资源指针
		foreach( Cache :: get( 'CACHE.THREAD' ) as $tData ) {
			if( is_resource( $tData[ 'FILE_HANDLE' ] ) ) {
				fclose( $tData[ 'FILE_HANDLE' ] );
			}
		}
		foreach( Cache :: get( 'MG.CURL_HANDLE' ) as $ch ) {
			curl_multi_remove_handle( Cache :: get( 'MG.MULTI_HANDLE' ), $ch );
		}
		curl_multi_close( Cache :: get( 'MG.MULTI_HANDLE' ) );
		Cache :: set( 'MG.CURL_HANDLE', null );

		// 任务被强制转为单线程，删除子线程，创建新的子进程
		if( true === Cache :: get( 'MG.THREAD_ONE' ) ) {
			Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> where( 'uid=' . Cache :: get( 'MG.MAIN_ID' ) ) -> delete();
			Cache :: set( 'MG.THREAD_ONE', false );
			$task = self :: task_main_select( Cache :: get( 'MG.MAIN_ID' ) );
			Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> replace( array(
				'uid' => Cache :: get( 'MG.MAIN_ID' ),
				'start' => 0,
				'end' => $task[ 0 ][ 'size' ] - 1,
				'pointer' => 0
			) );
			Cache :: set( 'CACHE', null );
			$this -> task_continue( Cache :: get( 'MG.MAIN_ID' ) );
			return false;
		}

		// 指针安全关闭，再次处理事件
		self :: events();

		// 任务完成判断，只查找此任务未完成的线程
		$thread_list = Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> where( 'uid=' . Cache :: get( 'MG.MAIN_ID' ) . ' AND thread_list.pointer < thread_list.end' ) -> order( 'level ASC' ) -> select();
		// 任务未完成！存在未完成线程并且状态允许
		if( count( $thread_list ) > 0 && self :: task_status() ) {
			Cache :: set( 'CACHE.THREAD', array() );
			$this -> task_continue( Cache :: get( 'MG.MAIN_ID' ) );
			return false;
		// 未完成的子线程不存在
		} else if( count( $thread_list ) < 1 ) {
			$task = Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( 'id=' . Cache :: get( 'MG.MAIN_ID' ) ) -> count();
			if( $task > 0 ) {
				// 任务完成！未完成的子线程不存在，但任务存在
				return $this -> task_complete( Cache :: get( 'MG.MAIN_ID' ) );
			}
		}

		return false;
	}

	/**
	 * 线程数据包返回处理
	 *
	 * @static
	 * @access public
	 * @param Resource $ch Curl Handle
	 * @param Data $data, 数据包，根据网络状况，大小可能约为 15000 Bytes
	 * @return Integer 必须返回数据包大小( Bytes )
	 */
	static public function multi_thread_callback( $ch, $data ) {

		$mac = MultiGet :: parseMac( $ch );

		// 本次返回的数据包大小
		$dataBytes = strlen( $data );

		// 当前线程追加缓存，写入文件后清空
		Cache :: set( 'CACHE.THREAD.' . $mac . '.DATA', $data, '.=' );
		// 当前线程已缓存大小，用于缓存控制写入，写入后清空
		Cache :: set( 'CACHE.THREAD.' . $mac . '.BYTES', $dataBytes, '+=' );
		// 当前线程已缓存大小，用于记录进度计算，记录后清空
		Cache :: set( 'CACHE.THREAD.' . $mac . '._BYTES', $dataBytes, '+=' );
		// 模拟指针移动，用于支持暂停断点续传
		Cache :: set( 'CACHE.THREAD.' . $mac . '.POINTER', $dataBytes, '+=' );

		// 空数据包，记录次数 + 1，必须是连续的空数据包，至少要有一个线程存活
		if( $dataBytes == 0 ) {
			Cache :: set( 'CACHE.NULL_DATA_NUM', 1, '+=' );
		} else {
			Cache :: set( 'CACHE.NULL_DATA_NUM', 0 );
		}

		return $dataBytes;
	}


	/**
	 * 线程进度更新
	 * 累计各线程缓存大小以及虚拟的指针位置
	 * 用于进度计算，断点续传等
	 *
	 * @access private
	 * @return Boolean
	 */
	private function multi_update() {

		$CACHE = Cache :: get( 'CACHE.THREAD' );
		if( count( $CACHE ) < 1 ) return false;

		/*
		 * 线程异常检查
		 * 可能某些服务器无视Headers: Range，导致每个线程返回相同的数据包
		 * 这里加一下验证，遇到这种情况强制使用单线程
		*/
		if( Cache :: get( 'MG.THREAD_NUM' ) > 1 && false == Cache :: get( 'MG.THREAD_CHECK' ) && thread_data_check( $CACHE ) ) {
			// 标记已经验证
			Cache :: set( 'MG.THREAD_CHECK', true );
			// 对比每个缓存的同一长度内容
			if( true === checkCache( $CACHE ) ) {
				// 强制停止，再次循环，并且改为单线程
				Cache :: set( 'MG.STATUS.START', false );
				Cache :: set( 'STATIC.REPEAT_NUM', 1, '-=' );
				Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( 'id=' . Cache :: get( 'MG.MAIN_ID' ) ) -> update( array(
					'thread' => 1,
					'@runtime' => 'runtime+' . Cache :: get( 'RUNTIME' )
				) );
				Cache :: set( 'RUNTIME', microtime( true ) );
				Cache :: set( 'MG.THREAD_ONE', true );
				echo 'THREAD ONE!', "\n";
				return false;
			}
		}

		$laterData = 0;
		foreach( $CACHE as $mac => $tData ) {
			if( empty( $tData[ '_BYTES' ] ) ) continue;
			Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> where( "id={$tData[ 'ID' ]}" ) -> update( array(
				'pointer' => $tData[ 'POINTER' ],
				'@data' => 'data+' . $tData[ '_BYTES' ]
			) );
			$laterData += $tData[ '_BYTES' ];
			Cache :: set( 'CACHE.THREAD.' . $mac . '._BYTES', 0 );
		}
		// 累计完成量及已用时间更新到主任务
		if( $laterData > 0 ) {
			Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( 'id=' . Cache :: get( 'MG.MAIN_ID' ) ) -> update( array(
				'@data' => 'data+' . $laterData,
				'@runtime' => 'runtime+' . round( microtime( true ) - Cache :: get( 'RUNTIME' ) )
			) );
			// 重置计时
			Cache :: set( 'RUNTIME', microtime( true ) );
		}
		return true;
	}

	/**
	 * 任务缓存控制，避免每次返回数据包都会对磁盘进行写入操作
	 * CONFIG.TASK_MEM_LIMIT ( MB )
	 * 线程缓存大小取决于同时可进行任务数和线程最大限制
	 * 需要额外的物理内存支持
	 *
	 * @access private
	 * @param Boolean $force 强制更新进度、写入缓存，用于突然中断或收尾
	 * @return Boolean
	 */
	private function task_cache( $force = false ) {

		$CACHE = Cache :: get( 'CACHE.THREAD' );

		// 用于计算进度和速度，越大的更新间隔(CONFIG.MULTI_UPDATE_INTERVAL)越能提高性能
		if( is_null( Cache :: get( 'CACHE_RUNTIME' ) ) ) {
			Cache :: set( 'CACHE_RUNTIME', microtime( true ) );
		} else if( $force || round( microtime( true ) - Cache :: get( 'CACHE_RUNTIME' ), 2 ) >= Cache :: get( 'CONFIG.MULTI_UPDATE_INTERVAL' ) ) {
			$this -> multi_update();
			Cache :: set( 'CACHE_RUNTIME', microtime( true ) );
		}


		if( ! is_array( $CACHE ) || count( $CACHE ) < 1 ) return false;

		// 线程缓存控制
		$thread_cache_limit = Cache :: get( 'MG.THREAD_CACHE_SIZE_LIMIT' );
		// 检查所有线程缓存
		foreach( $CACHE as $mac => $tData ) {
			// 当线程结束或缓存超过大小限制，写入缓存，忽略空数据
			if( $force == true || ( ! empty( $tData[ 'BYTES' ] ) && $tData[ 'BYTES' ] >= $thread_cache_limit ) ) {
				if( ! empty( $tData[ 'DATA' ] ) ) {
					fwrite( $tData[ 'FILE_HANDLE' ], $tData[ 'DATA' ] );
				}
				// 重置当前线程缓存
				Cache :: set( 'CACHE.THREAD.' . $mac . '.DATA', null );
				Cache :: set( 'CACHE.THREAD.' . $mac . '.BYTES', 0 );
			}
		}
		return self :: task_status();
	}

	/**
	 * 更新任务状态
	 *
	 * @static
	 * @access public
	 * @return Boolean
	 */
	static public function task_status() {
		$task = self :: task_main_select( Cache :: get( 'MG.MAIN_ID' ) );
		if( count( $task ) < 1 || $task[ 0 ][ 'status' ] != TASK_START ) {
			Cache :: set( 'MG.STATUS.START', false );
			return false;
		}
		return true;
	}

	/**
	 * 获取任务信息
	 *
	 * @static
	 * @access public
	 * @param String $url
	 * @return Boolean | Array
	 */
	static public function task_header( $url ) {
		$info = get_headers( $url, true );
		if( ! is_array( $info ) || empty( $info[ 'Content-Type' ] ) || empty( $info[ 'Content-Length' ] ) ) {
			return false;
		} else if( empty( $info[ 'Content-Type' ] ) ) {
			return false;
		} else if( is_string( $info[ 'Content-Type' ] ) && strpos( $info[ 'Content-Type' ], 'text/html' ) !== false ) {
			return false;
		} else if( is_array( $info[ 'Content-Type' ] ) && ! in_array( 'application/octet-stream', $info[ 'Content-Type' ]  ) ) {
			return false;
		}
		$http = 0;
		if( isset( $info[ 'Accept-Ranges' ] ) || floatval( substr( $info[ 0 ], strpos( $info[ 0 ], '/' ) + 1, 3 ) ) > 1.0 ) {
			$http = 1;
		}
		if( is_array( $info[ 'Content-Length' ] ) ) {
			do{
				$size = array_pop( $info[ 'Content-Length' ] );
			} while( $size == 0 && count( $info[ 'Content-Length' ] ) );
		} else {
			$size = $info[ 'Content-Length' ];
		}
		return array( 'size' => $size, 'http' => $http );
	}

	/**
	 * 删除一个任务
	 *
	 * @static
	 * @access public
	 * @param String | Integer $urlORid
	 * @param Boolean $rm_file 同时删除文件
	 * @return Boolean
	 */
	static public function task_remove( $urlORid, $rm_file = false ) {

		$where = self :: task_main_select( $urlORid, true );
		$task = Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( $where ) -> select();
		if( count( $task ) < 1 ) return false;
		if( $rm_file == true ) {
			if( file_exists( $task[ 0 ][ 'file_name' ] . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ) ) ) {
				@unlink( $task[ 0 ][ 'file_name' ] . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ) );
			} else if( file_exists( $task[ 0 ][ 'file_name' ] ) ) {
				@unlink( $task[ 0 ][ 'file_name' ] );
			}
		}
		// 删除任务及子线程
		Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> where( "uid={$task[ 0 ][ 'id' ]}" ) -> delete();
		Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( $where ) -> delete();
		return true;

	}

	/**
	 * 队列等待
	 *
	 * @static
	 * @access public
	 * @param String | Integer $urlORid
	 * @return Boolean | Integer
	 */
	static public function task_standby( $urlORid ) {
		$where = self :: task_main_select( $urlORid, true );
		$task = self :: task_main_select( $urlORid );
		if( count( $task ) < 1 || $task[ 0 ][ 'status' ] == TASK_COMPLETE ) return false;
		return Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( $where ) -> update( array( 'status' => TASK_STANDBY ) );
	}

	/**
	 * 任务暂停
	 *
	 * @static
	 * @access public
	 * @param String | Integer $urlORid
	 * @return Boolean | Integer
	 */
	static public function task_pause( $urlORid ) {
		$where = self :: task_main_select( $urlORid, true );
		$task = self :: task_main_select( $urlORid );
		// 状态大于开始的(1~3)不可能转换为暂停
		if( count( $task ) < 1 || $task[ 0 ][ 'status' ] > TASK_START ) return false;
		Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( $where ) -> update( array(
			'status' => TASK_PAUSE,
			'@runtime' => 'runtime+' . round( microtime( true ) - Cache :: get( 'RUNTIME' ) )
		) );
	}

	/**
	 * 任务完成
	 *
	 * @access private
	 * @param String | Integer $urlORid
	 * @return Boolean
	 */
	private function task_complete( $urlORid ) {
		$where = self :: task_main_select( $urlORid, true );
		$task = self :: task_main_select( $urlORid );
		Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( $where ) -> update( array(
			'status' => TASK_COMPLETE,
			'update_time' => time(),
			'level' => 0,
			'@runtime' => 'runtime+' . round( microtime( true ) - Cache :: get( 'RUNTIME' ) )
		) );
		// 删除子线程
		Cache :: get( 'SQLITE' ) -> table( 'thread_list' ) -> where( "uid={$task[ 0 ][ 'id' ]}" ) -> delete();
		// 去掉临时文件的后缀
		if( file_exists( $task[ 0 ][ 'file_name' ] . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ) ) ) {
			rename( $task[ 0 ][ 'file_name' ] . Cache :: get( 'CONFIG.TASK_TEMP_EXT' ), $task[ 0 ][ 'file_name' ] );
		}
		return false;
	}

	/**
	 * 任务失败
	 *
	 * @access private
	 * @param String | Integer $urlORid
	 * @return Boolean | Integer
	 */
	private function task_failure( $urlORid ) {
		$where = self :: task_main_select( $urlORid, true );
		Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( $where ) -> update( array(
			'status' => TASK_FAILURE,
			'level' => 0,
			'@runtime' => 'runtime+' . round( microtime( true ) - Cache :: get( 'RUNTIME' ) ),
			'update_time' => time()
		) );
		return Cache :: get( 'SQLITE' ) -> numRows;
	}

	/**
	 * 用url或id查询一个主任务
	 *
	 * @static
	 * @access public
	 * @param String | Integer $urlORid
	 * @param Boolean $return 返回用于查询的条件字符串
	 * @return Array | String
	 */
	static public function task_main_select( $urlORid, $return = false ) {
		if( is_numeric( $urlORid ) && is_numeric( $urlORid ) == $urlORid ) {
			$where = "id={$urlORid}";
		} else {
			$where = "url='{$urlORid}'";
		}
		if( $return == true ) return $where;
		return Cache :: get( 'SQLITE' ) -> table( 'task_list' ) -> where( $where ) -> select();
	}

	/**
	 * 强制转化一个资源到字符串, 用于定位线程
	 *
	 * @static
	 * @access public
	 * @param Resource $ch
	 * @return String
	 */
	static public function parseMac( $ch ) {
		if( ! is_resource( $ch ) ) return false;
		return substr( md5( ( string ) $ch ), 17 );
	}

}

?>