<?PHP
/**
 * PHP Version 5.2 +
 * Extension php_sqlite.dll, php_sqlite.so
 * Copyright (C) 2010 All Rights Reserved.
 *
 * @Application : SQLite.class.php
 * @Encode : 65001 Unicode ( UTF-8 NO BOM )
 * @Description : SQLite 数据库操作类，非PDO版本
 * @Author : HK <hkyyang@qq.com>
 */
class SQLite {

	// SQLite 版本号
	public $dbVersion;
	// SQLite 连接编码
	public $dbEncode;
	// 最后执行的SQL语句
	public $queryStr = '';
	// 执行语句的影响行数
	public $numRows = 0;
	// 最后插入数据库的id
	public $lastInsertID = null;
	// 当前数据库名称
	public $dbName;
	// 当前数据库绝对路径
	public $dbFile;
	// 错误信息
	public $error;

	// 当前打开的数据库连接资源
	protected $dbHandle = null;
	// 最后执行SQL的资源
	protected $queryID = null;
	// 链式操作参数
	protected $options = array();

	// 是否持久连接
	private $pconnect;
	// 数据库连接成功标记
	private $connected = false;
	// 数据库连接超时毫秒
	private $timeout = 10000;


	// SELECT 语句模板
	const __TPL_SELECT__ = 'SELECT%DISTINCT% %FIELDS% FROM %TABLE%%JOIN%%WHERE%%GROUP%%HAVING%%ORDER%%LIMIT%';
	// DELETE 语句模板
	const __TPL_DELETE__ = 'DELETE FROM %TABLE%%WHERE%';
	// UPDATE 语句模板
	const __TPL_UPDATE__ = 'UPDATE %TABLE% SET %VALUE%%WHERE%';
	// INSERT 语句模板
	const __TPL_INSERT__ = 'INSERT INTO %TABLE% ( %KEY% ) VALUES( %VALUE% )';
	// REPLACE 语句模板
	const __TPL_REPLACE__ = 'REPLACE INTO %TABLE% ( %KEY% ) VALUES( %VALUE% )';

	/**
	 *  构造方法，SQLite初始化
	 *
	 * @access public
	 * @param String, $dbName 数据库名(相对), 或数据库文件绝对物理地址
	 * @param Boolean, $autoCreate 当数据库不存在时，创建数据库
	 * @param Boolean, $pconnect 打开一个到 SQLite 的持久连接
	 * @return void
	 */
    public function __construct( $dbName = null, $autoCreate = true, $pconnect = false ){

		if( is_null( $dbName ) ) {
			$this -> error = 'Error: Missing argument "database file name".';
        } else if ( intval( phpversion() ) < 5 ) {
            $this -> error = 'Error: Need a higher version.';
        } else if( ! extension_loaded( 'sqlite' ) ) {
			$this -> error = "Error: Unable to load extension \"php_sqlite.". PHP_SHLIB_SUFFIX ."\".";
		}
		if( empty( $this -> error ) && ! file_exists( $dbName ) && ! $autoCreate ) {
			$this -> error = "Error: Unable to open database '{$dbName}'.";
		}

		$this -> errorHalt();

		$this -> dbName = $dbName;
		$this -> dbFile = $dbName;
		$this -> dbVersion = sqlite_libversion();
		$this -> dbEncode = sqlite_libencoding();
		$this -> pconnect = $pconnect ? true : false;

    }

	/**
	 *  SQLite SELECT 查询数据库
	 *
	 * @access public
	 * @param String, $tbName 表名
	 * @return Array 返回结果数组
	 */
	public function select( $tbName = null ) {
		$sql = $this -> parseSQL( 'SELECT' );
		if( strpos( $sql, '%TABLE%' ) !== false ) {
			$sql = str_replace( '%TABLE%', $tbName, $sql );
		}
		$this -> queryStr = $sql;
		return $this -> query();
	}

	/**
	 *  SQLite UPDATE 更新数据库
	 *
	 * @access public
	 * @param String, $tbName 表名
	 * @param Array, $value 内容数组, array( 字段名 => 值 ), 如果不想值被强加引号，在字段名后加@
	 * @return Integer 影响行数
	 */
	public function update( $tbName = null, $value = array() ) {
		$sql = $this -> parseSQL( 'UPDATE' );
		if( is_array( $tbName ) ) {
			$value = $tbName;
		}
		if( strpos( $sql, '%TABLE%' ) !== false && is_string( $tbName ) ) {
			$sql = str_replace( '%TABLE%', $tbName, $sql );
		}
		$set = '';
		foreach( $value as $k => $v ) {
			if( strpos( $k, '@' ) > -1 ) {
				$set .= str_replace( '@', '', $k ) . '=' . $this -> escape_string( $v ) . ', ';
			} else {
				$set .= $k . '=' . "'" . $this -> escape_string( $v ) . "', ";
			}
		}
		$set = rtrim( $set, ', ' );
		$this -> queryStr = str_replace( '%VALUE%', $set, $sql );
		$this -> execute();
		return $this -> numRows;
	}

	/**
	 *  SQLite REPLACE INTO 向数据表中插入并覆盖数据
	 *
	 * @access public
	 * @param String, $tbName 表名
	 * @param Array, $value 插入的内容数组,支持二维数组
	 * @return Integer 返回插入数量
	 */
	public function replace( $tbName = null, $value = array() ) {
		return $this -> insert( $tbName, $value, true );
	}

	/**
	 *  SQLite INSERT INTO 向数据表中插入数据
	 *
	 * @access public
	 * @param String, $tbName 表名
	 * @param Array, $value 插入的内容数组,支持二维数组
	 * @return Integer 返回插入数量
	 */
	public function insert( $tbName = null, $value = array(), $replace = false ) {
		$cmd = $replace ? 'REPLACE' : 'INSERT';
		$sql = $this -> parseSQL( $cmd );
		if( is_array( $tbName ) ) {
			$value = $tbName;
		}
		if( strpos( $sql, '%TABLE%' ) !== false && is_string( $tbName ) ) {
			$sql = str_replace( '%TABLE%', $tbName, $sql );
		}
		$sql_str = $val = $key = '';
		$sql_arr = array();
		if( count( $value ) < 1 ) $this -> errorHalt( 'Error: Table name not found.' );
		if( ! empty( $value[ 0 ] ) && is_array( $value[ 0 ] ) ) {
			$key = '';
			for( $i = 0; $i < count( $value ); $i ++ ) {
				$val = '';
				$sql_str = $sql;
				foreach( $value[ $i ] as $k => $v ) {
					if( $i == 0 ) $key .= $k . ', ';
					if( is_int( $v ) ) {
						$val .= $v . ', ';
					} else {
						$val .= "'" . $this -> escape_string( $v ) . "', ";
					}
				}
				$key = rtrim( $key, ', ' );
				$val = rtrim( $val, ', ' );
				$sql_arr[] = str_replace(
					array( '%KEY%', '%VALUE%' ),
					array( $key, $val ),
					$sql
				);
			}
		} else {
			foreach( $value as $k => $v ) {
				$key .= $k . ', ';
				if( is_int( $v ) ) {
					$val .= $v . ', ';
				} else {
					$val .= "'" . $this -> escape_string( $v ) . "', ";
				}
			}
			$key = rtrim( $key, ', ' );
			$val = rtrim( $val, ', ' );
			$this -> queryStr = str_replace(
				array( '%KEY%', '%VALUE%' ),
				array( $key, $val ),
				$sql
			);
		}

		if( count( $sql_arr ) ) {
			$num_rows = 0;
			for( $i = 0; $i < count( $sql_arr ); $i ++ ) {
				$this -> queryStr = $sql_arr[ $i ];
				$num_rows += $this -> execute();
			}
			return $num_rows;
		}

		$this -> execute();
		return $this -> numRows;
	}

	/**
	 *  SQLite DELETE语句
	 *
	 * @access public
	 * @param String, $tbName 表名
	 * @return Integer
	 */
	public function delete( $tbName = null ) {
		$sql = $this -> parseSQL( 'DELETE' );
		if( strpos( $sql, '%TABLE%' ) !== false ) {
			$sql = str_replace( '%TABLE%', $tbName, $sql );
		}
		$this -> queryStr = $sql;
		$this -> execute();
		// 成功删除后，整理数据库释放空间，需要 SQLite 2.8.1 +
		if( $this -> numRows > 0 && intval( substr( str_replace( '.', '', $this -> dbVersion ), 0, 3 ) ) >= 281 ) {
			$this -> opquery( 'VACUUM ' );
		}
		return $this -> numRows;
	}

	/**
	 *  SQLite 实现 TRUNCATE TABLE
	 *
	 * @access public
	 * @param String, $tbName 表名
	 * @return Boolean
	 */
	public function truncate( $tbName = null ) {
		$tbName = $tbName ? $tbName : $this -> parseGet( 'table' );
		if( empty( $tbName ) ) $this -> errorHalt( 'Error: Table name not found.' );
		$info = $this -> table( 'sqlite_master' ) -> where( "type='table' AND name='{$tbName}'" ) -> select();
		if( ! empty( $info ) && $this -> opquery( 'DROP TABLE ' . $tbName ) ) {
			if( $this -> opquery( $info[ 0 ][ 'sql' ] ) && intval( substr( str_replace( '.', '', $this -> dbVersion ), 0, 3 ) ) >= 281 ) {
				return $this -> opquery( 'VACUUM ' );
			}
		}
		return false;
	}

	/**
	 *  SQLite info
	 * 简陋的DB信息，仅供个人调试
	 *
	 * @access public
	 * @param Boolean, $show 显示信息
	 * @return void | Array
	 */
	public function info( $show = true ) {

		$info = $this -> table( 'sqlite_master' )  -> where( "type='table'" ) -> select();
		$tables = count( $info );
		$size = sizeCount( filesize( $this -> dbFile ) );
		if( $show === false ) return $info;
		echo "<style>table{border-collapse:collapse;border: 1px solid #EFEFEF; float:left; margin-top:10px; fnot-family:Arial;} table th{background:#F9F9F9; border:1px solid #EFEFEF; font-weight:normal; height:30px; line-height:30px; font-size:12px;} table tr td{ font-size:12px; color:#666; height:20px; line-height:20px; text-align:left; padding-left:1%; border:1px solid #EFEFEF; overflow:hidden;} .tab{margin-left:10px;} .info{color:#666; font-size:10px; line-height:15px; width:300px; border:1px solid #EFEFEF;} .info .head{display:block;background:#EFEFEF; width:290px; color:#666; padding:0 5px 0 5px;} .info div{padding-left:5px; color:#999;}</style>\n<div class=\"info\"><span class=\"head\">SQLite</span><div>SQLite Version: {$this -> dbVersion}</div><div>Libencode: {$this -> dbEncode}</div><div>DB Name: {$this -> dbName}</div><div>DB File: {$this -> dbFile}</div><div>DB Size: " . $size[ 'size' ] , ' ', $size[ 'full_name' ] . "</div><div>Tables: {$tables}</div><div>PHP Version: " . phpversion() . "</div></div>\n";
		echo "<table>\n";
		for( $i = 0; $i < count( $info ); $i ++ ) {
			$num = $this -> table( $info[ $i ][ 'name' ] ) -> count();
			if( $i < 1 ) echo "<tr><th width=\"120\">Type</th><th width=\"150\">Name</th><th width=\"150\">Tbl Name</th><th width=\"50\" style=\"font-size:9px;\">Root Page</th><th width=\"300\">SQL</th></tr>\n";
			echo "<tr><td>{$info[ $i ][ 'type' ]}</td><td>{$info[ $i ][ 'name' ]} ({$num})</td><td>{$info[ $i ][ 'tbl_name' ]}</td><td>{$info[ $i ][ 'rootpage' ]}</td><td>" . highlight_string( $info[ $i ][ 'sql'], true ) . "</td></tr>\n";
		}
		echo "<table>\n";
	}

	/**
	 *  SQLite表创建
	 *
	 * @access public
	 * @param String, $tbName 表名
	 * @param Array, $fields 字段名(Key)及类型(Value)
	 * @param Array, $index  索引
	 * @return Boolean
	 */
	public function create( $tbName, $fields = array(), $index = array() ) {
		$sql = "CREATE TABLE $tbName (\n";
		foreach( $fields as $name => $type ) {
			$sql .= "\t" . $name . ' ' . $type . ",\n";
		}
		if( count( $index ) ) {
			foreach( $index as $k => $v ) {
				$sql .= "\t$k ( $v ),\n";
			}
		}
		$sql = trim( $sql, ",\n" ) . "\n)";
		$this -> queryStr = $sql;
		return $this -> opquery();
	}

	/**
	 *  执行SQL,只返回成功结果
	 *
	 * @access public
	 * @param String, $sql
	 * @return Boolean
	 */
	public function opquery( $sql = '' ) {
		$sql = $sql ? $sql : $this -> queryStr;
		$this -> queryStr = $sql;
		$this -> connect();
		$this -> queryID = @sqlite_query( $this -> dbHandle, $sql, SQLITE_ASSOC, $this -> error );
		if( ! $this -> queryID ) {
			return false;
		} else {
			return true;
		}
	}


	/**
	 *  统计查询
	 *
	 * @access private
	 * @param String, $str
	 * @return Integer
	 */
	private function getField( $str, $alias ) {
		$sql = $this -> parseSQL( 'COUNT' );
		$this -> queryStr = str_replace( '%FIELDS%', $str, $sql );
		$info = $this -> query();
		if( ! empty( $info[ 0 ][ $alias ] ) ) return $info[ 0 ][ $alias ];
		return 0;
	}

	/**
	 *  SQL语句构造控制
	 *
	 * @access private
	 * @param $
	 * @return String
	 */
	private function parseSQL( $tpl ) {
		$sql = '';
		switch( strtoupper( $tpl ) ) {

			case 'SELECT' :
				$sql = self :: __TPL_SELECT__;
				$sql = str_replace(
						array( '%TABLE%', '%DISTINCT%', '%FIELDS%', '%JOIN%', '%WHERE%', '%GROUP%', '%HAVING%', '%ORDER%', '%LIMIT%' ),
						array(
							$this -> parseGet( 'table' ),
							$this -> parseGet( 'distinct' ),
							$this -> parseGet( 'field' ),
							$this -> parseGet( 'join' ),
							$this -> parseGet( 'where' ),
							$this -> parseGet( 'group' ),
							$this -> parseGet( 'having' ),
							$this -> parseGet( 'order' ),
							$this -> parseGet( 'limit' )
						), $sql );
				break;

			case 'DELETE' :
				$sql = self :: __TPL_DELETE__;
				$sql = str_replace(
						array( '%TABLE%', '%WHERE%' ),
						array(
							$this -> parseGet( 'table' ),
							$this -> parseGet( 'where' )
				), $sql );
				break;

			case 'REPLACE' :
				$sql = self :: __TPL_REPLACE__;
				$sql = str_replace(
						array( '%TABLE%', '%WHERE%' ),
						array(
							$this -> parseGet( 'table' ),
							$this -> parseGet( 'where' ),
						), $sql );
				break;

			case 'INSERT' :
				$sql = self :: __TPL_INSERT__;
				$sql = str_replace(
						array( '%TABLE%', '%WHERE%' ),
						array(
							$this -> parseGet( 'table' ),
							$this -> parseGet( 'where' ),
						), $sql );
				break;

			case 'UPDATE' :
				$sql = self :: __TPL_UPDATE__;
				$sql = str_replace(
						array( '%TABLE%', '%WHERE%' ),
						array(
							$this -> parseGet( 'table' ),
							$this -> parseGet( 'where' )
						), $sql );
				break;
			case 'COUNT' :
				$sql = self :: __TPL_SELECT__;
				$sql = str_replace(
						array( '%TABLE%', '%DISTINCT%', '%JOIN%', '%WHERE%', '%GROUP%', '%HAVING%', '%ORDER%', '%LIMIT%' ),
						array(
							$this -> parseGet( 'table' ),
							$this -> parseGet( 'distinct' ),
							$this -> parseGet( 'join' ),
							$this -> parseGet( 'where' ),
							$this -> parseGet( 'group' ),
							$this -> parseGet( 'having' ),
							$this -> parseGet( 'order' ),
							$this -> parseGet( 'limit' )
						), $sql );
				break;
		}
		unset( $this -> options );
		return $sql;
	}

	/**
	 *  获取链式操作中的参数
	 *
	 * @access private
	 * @param String, $type
	 * @return String
	 */
	private function parseGet( $type ) {
		$options = & $this -> options;
		$type = strtolower( $type );
		$value = '';
		switch( $type ) {
			case 'table' : $value = empty( $options[ 'table' ] ) ? '%TABLE%' : $options[ 'table' ]; break;
			case 'distinct' : $value = empty( $options[ 'distinct' ] ) ? '' : ' DISTINCT '; break;
			case 'field' : $value = empty( $options[ 'field' ] ) ? '*' : $options[ 'field' ]; break;
			case 'join' : $value = empty( $options[ 'join' ] ) ? '' : ' LEFT JOIN ' . $options[ 'join' ]; break;
			case 'where' : $value = empty( $options[ 'where' ] ) ? '' : ' WHERE ' . $options[ 'where' ]; break;
			case 'group' : $value = empty( $options[ 'group' ] ) ? '' : ' GROUP BY ' . $options[ 'group' ]; break;
			case 'having' : $value = empty( $options[ 'having' ] ) ? '' : ' HAVING ' . $options[ 'having' ]; break;
			case 'order' : $value = empty( $options[ 'order' ] ) ? '' : ' ORDER BY ' . $options[ 'order' ]; break;
			case 'limit' : $value = empty( $options[ 'limit' ] ) ? '' : ' LIMIT ' . $options[ 'limit' ]; break;
		}
		if( isset( $options[ $type ] ) ) unset( $options[ $type ] );
		return $value;
	}

	/**
	 *  连接SQLite数据库
	 *
	 * @access private
	 * @return Source
	 */
	private function connect() {

		if( $this -> connected == true ) {
			return $this -> dbHandle;
		}

		$openfunc = $this -> pconnect ? 'sqlite_popen' : 'sqlite_open';
		$this -> dbHandle = $openfunc( $this -> dbName, 0666, $this -> error );

		if( ! $this -> dbHandle ) {
			$this -> errorHalt( $this -> error );
		}
		$this -> connected = true;
		sqlite_busy_timeout( $this -> dbHandle, $this -> timeout );
		return $this -> dbHandle;
	}

	/**
	 *  关闭SQLite连接
	 *
	 * @access private
	 * @return void
	 */
	private function close() {
		if( $this -> dbHandle ) {
			sqlite_close( $this -> dbHandle );
		}
		$this -> dbHandle = null;
	}

	/**
	 *  用于执行INSERT, UPDATE等
	 *
	 * @access private
	 * @return Integer
	 */
	private function execute() {
		$this -> connect();
		$this -> queryID = @sqlite_query( $this -> dbHandle, $this -> queryStr, SQLITE_ASSOC, $this -> error );
		if( false === $this -> queryID ) {
			$this -> errorHalt( $this -> error );
		} else {
			$this -> numRows = sqlite_changes( $this -> dbHandle );
			$this -> lastInsertID = sqlite_last_insert_rowid( $this -> dbHandle );
		}
		return $this -> numRows;
	}

	/**
	 *  用于执行SELECT
	 *
	 * @access private
	 * @return Array
	 */
	private function query() {
		$this -> connect();
		$this -> queryID = @sqlite_query( $this -> dbHandle, $this -> queryStr, SQLITE_ASSOC, $this -> error );
		if( false === $this -> queryID ) {
			$this -> errorHalt( $this -> error );
		} else {
			$this -> numRows = sqlite_num_rows( $this -> queryID );
		}
		return $this -> fetch_all();
	}

	/**
	 *  从结果集中取得数组
	 *
	 * @access private
	 * @param String, $string
	 * @return String
	 */
	private function fetch_all() {
        $result = array();
        if( $this -> dbHandle && $this -> numRows > 0 ) {
            while( $row = sqlite_fetch_array( $this->queryID ) ) {
                $result[] = $row;
            }
            sqlite_seek( $this -> queryID, 0 );
        }
        return $result;
	}

	/**
	 *  转义一个字符串用于 sqlite_query
	 *
	 * @access private
	 * @param String, $string
	 * @return String
	 */
	private function escape_string( $string ) {
		return sqlite_escape_string( $string );
	}

	/**
	 * 检查错误输出并停止
	 *
	 * @access private
	 * @param String, $error 错误信息
	 * @return void
	 */
	private function errorHalt( $error = null ) {
		$error = $error ? $error : $this -> error;
		if( empty( $error ) && $this -> dbHandle ) {
			$error = sqlite_last_error( $this -> dbHandle );
			if( $error ) $error = sqlite_error_string( $error );
		}
		if( ! empty( $error ) ) {
			echo '<div style="color:#900;">Error: ' . ltrim( $error, 'Error: ' ) . '</div>';
			if( ! empty( $this -> queryStr ) ) echo '<div style="color:#900;">SQL: '. $this -> queryStr .'</div>';
			exit;
		}
	}

	/**
	 * 魔术方法__call, 实现链式操作
	 *
	 * @access public
	 * @param $
	 * @return
	 */
	public function __call( $method, $args ) {
		$method = strtolower( $method );
		if( in_array( $method, array( 'field', 'table', 'where', 'order', 'limit', 'page', 'having', 'group', 'distinct' ), true ) ) {
			$this -> options[ $method ] = $args[ 0 ];
			return $this;
		} else if( in_array( $method, array( 'count', 'sum', 'min', 'max', 'avg' ), true ) ) {
			$field = empty( $args[ 0 ] ) ? '*' : $args[ 0 ];
			$alias = empty( $args[ 1 ] ) ? $method . '_nums' : $args[ 1 ];
			return $this -> getField( strtoupper( $method ) . '( ' . $field . ' )' . ' AS ' . $alias, $alias );
		} else {
			$this -> errorHalt( 'Error: ' . __CLASS__ . ':' . $method . ' not found.' );
		}
		return false;
	}

	/**
	 * 析构方法__destruct, 关闭SQLite连接
	 *
	 * @access public
	 * @param $
	 * @return
	 */
	public function __destruct() {
		$this -> close();
	}

	// @ toString
	public function __toString() {
		return 'Object: SQLite ' . $this -> dbVersion;
	}

}
?>